'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.CrioObject = exports.CrioArray = exports.mergeCrios = exports.mergeArrays = exports.getCrioedValue = undefined;

var _CRIO_ARRAY_PROTOTYPE, _CRIO_OBJECT_PROTOTYP;

var _hashIt = require('hash-it');

var _hashIt2 = _interopRequireDefault(_hashIt);

var _forEach = require('lodash/forEach');

var _forEach2 = _interopRequireDefault(_forEach);

var _forEachRight = require('lodash/forEachRight');

var _forEachRight2 = _interopRequireDefault(_forEachRight);

var _isArray = require('lodash/isArray');

var _isArray2 = _interopRequireDefault(_isArray);

var _isEqual = require('lodash/isEqual');

var _isEqual2 = _interopRequireDefault(_isEqual);

var _isPlainObject = require('lodash/isPlainObject');

var _isPlainObject2 = _interopRequireDefault(_isPlainObject);

var _isUndefined = require('lodash/isUndefined');

var _isUndefined2 = _interopRequireDefault(_isUndefined);

var _constants = require('./utils/constants');

var _crio = require('./utils/crio');

var _loops = require('./utils/loops');

var _is = require('./utils/is');

var _stringify = require('./utils/stringify');

var _stringify2 = _interopRequireDefault(_stringify);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

function _toArray(arr) { return Array.isArray(arr) ? arr : Array.from(arr); }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } } // external dependencies


// utils


/**
 * get the crioed value if it is an array or object,
 * else return the value itself
 *
 * @param {*} value
 * @returns {*}
 */
var getCrioedValue = function getCrioedValue(value) {
  if ((0, _is.isCrio)(value) || (0, _is.isReactElement)(value)) {
    return value;
  }

  if ((0, _isArray2.default)(value)) {
    return new CrioArray(value);
  }

  if ((0, _isPlainObject2.default)(value)) {
    return new CrioObject(value);
  }

  return value;
};

/**
 * shallowly merge source arrays into target array
 * 
 * @param {Array<*>} target
 * @param {Array<array>} sources
 * @returns {Array<*>}
 */
var mergeArrays = function mergeArrays(target, sources) {
  return sources.reduce(function (plainObject, source) {
    if ((0, _isArray2.default)(source)) {
      (0, _forEach2.default)(source, function (value, index) {
        plainObject[index] = getCrioedValue(value);
      });
    }

    return plainObject;
  }, [].concat(_toConsumableArray(target)));
};

/**
 * shallowly merge sources into target
 *
 * @param {CrioArray|CrioObject} target
 * @param {Array<array|Object>} sources
 * @returns {CrioArray|CrioObject}
 */
var mergeCrios = function mergeCrios(target) {
  for (var _len = arguments.length, sources = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    sources[_key - 1] = arguments[_key];
  }

  if (!sources.length) {
    return target;
  }

  var mergeFunction = (0, _is.isCrioObject)(target) || (0, _isPlainObject2.default)(target) ? _loops.mergeObjects : mergeArrays;

  return (0, _crio.getSameCrioIfUnchanged)(target, mergeFunction(target, sources));
};

var Crio =
/**
 * create based Crio class with a null prototype that will assign
 * the values passed to itself
 *
 * @param {Array<*>|Object} object
 * @param {string} hashCode=hashIt(object)
 * @return {CrioArray|CrioObject}
 */
function Crio(object) {
  var _this = this;

  var hashCode = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : (0, _hashIt2.default)(object);

  _classCallCheck(this, Crio);

  if ((0, _is.isCrio)(object)) {
    return object;
  }

  var length = 0;

  (0, _forEach2.default)(object, function (value, key) {
    _this[key] = getCrioedValue(value);

    length++;
  });

  _constants.OBJECT.defineProperties(this, _defineProperty({
    length: {
      enumerable: false,
      value: length
    }

  }, _constants.CRIO_HASH_CODE, {
    enumerable: false,
    value: hashCode
  }));

  return (0, _crio.freezeIfNotProduction)(this);
};

var CRIO_PROTOTYPE = {
  /**
   * return an empty crio
   *
   * @returns {CrioArray|CrioObject}
   */
  clear: function clear() {
    if (!this.length) {
      return this;
    }

    var plainObject = (0, _crio.getPlainObject)(this);

    return new this.constructor(plainObject);
  },


  /**
   * reduce the Crio to only having values that are truthy
   *
   * @returns {CrioArray|CrioObject}
   */
  compact: function compact() {
    var compactedCrio = this.filter(function (value) {
      return !!value;
    });

    return this.equals(compactedCrio) ? this : compactedCrio;
  },


  constructor: Crio,

  /**
   * remove key from this
   *
   * @param {string|number} key
   * @returns {CrioArray|CrioObject}
   */
  delete: function _delete(key) {
    var _this2 = this;

    if (!this.has(key)) {
      return this;
    }

    var plainObject = (0, _crio.getPlainObject)(this),
        isThisArray = (0, _isArray2.default)(plainObject);

    (0, _forEach2.default)(this.keys(), function (currentKey) {
      if (currentKey !== key) {
        if (isThisArray) {
          plainObject.push(_this2[currentKey]);
        } else {
          plainObject[currentKey] = _this2[currentKey];
        }
      }
    });

    return (0, _crio.getSameCrioIfUnchanged)(this, plainObject);
  },


  /**
   * remove deeply-nested key from this
   *
   * @param {Array<string|number>} keys
   * @returns {CrioArray|CrioObject}
   */
  deleteIn: function deleteIn(keys) {
    var _this3 = this;

    (0, _crio.throwTypeErrorIfKeysInvalid)(keys);

    var length = keys.length;

    if (length === 1) {
      return this.delete(keys[0]);
    }

    if (!length) {
      return this;
    }

    var key = keys.shift();

    if (!this.has(key)) {
      return this;
    }

    var plainObject = (0, _crio.getPlainObject)(this),
        isTargetKey = false;

    this.forEach(function (currentValue, currentKey) {
      isTargetKey = currentKey === key;
      currentValue = _this3[currentKey];

      if (isTargetKey) {
        if ((0, _is.isCrio)(currentValue)) {
          plainObject[currentKey] = currentValue.deleteIn(keys);
        }
      } else {
        plainObject[currentKey] = currentValue;
      }
    });

    return (0, _crio.getSameCrioIfUnchanged)(this, plainObject);
  },


  /**
   * return an array of [key, value] pairs for this
   *
   * @returns {Array<array>}
   */
  entries: function entries() {
    return (0, _constants.OBJECT_ENTRIES)(this);
  },


  /**
   * determine if object passed is equal in value to this
   *
   * @param {CrioArray|CrioObject} object
   * @returns {boolean}
   */
  equals: function equals(object) {
    if (!(0, _is.isCrio)(object)) {
      return false;
    }

    return this[_constants.CRIO_TYPE] === object[_constants.CRIO_TYPE] && this[_constants.CRIO_HASH_CODE] === object[_constants.CRIO_HASH_CODE];
  },


  /**
   * get the value that matches at key
   *
   * @param {string|number} key
   * @returns {*}
   */
  get: function get(key) {
    return this[key];
  },


  /**
   * get the value that matches at the deeply nested location from keys
   *
   * @param {Array<string|number>} keys
   * @returns {*}
   */
  getIn: function getIn(keys) {
    (0, _crio.throwTypeErrorIfKeysInvalid)(keys);

    var length = keys.length;

    if (length === 1) {
      return this[keys[0]];
    }

    if (!length) {
      return this;
    }

    var currentObject = this,
        index = -1,
        key = void 0;

    while (++index < length) {
      key = keys[index];

      if ((0, _isUndefined2.default)(currentObject[key])) {
        return undefined;
      }

      if (index === length - 1) {
        return currentObject[key];
      }

      currentObject = currentObject[key];
    }

    return undefined;
  },


  /**
   * does this have the property passed
   *
   * @param {number|string} key
   * @returns {boolean}
   */
  has: function has(key) {
    return this.hasOwnProperty(key);
  },


  /**
   * does this have the property deeply nested
   *
   * @param {Array<number|string>} keys
   * @returns {boolean}
   */
  hasIn: function hasIn(keys) {
    (0, _crio.throwTypeErrorIfKeysInvalid)(keys);

    var length = keys.length;

    if (keys.length === 1) {
      return this.has(keys[0]);
    }

    if (!length) {
      return false;
    }

    var _keys = _toArray(keys);

    var key = _keys[0];

    var restOfKeys = _keys.slice(1);

    var target = this[key];

    if ((0, _is.isCrio)(target)) {
      return target.hasIn(restOfKeys);
    }

    return false;
  },


  /**
   * does this have the property passed
   *
   * @param {number|string} key
   * @returns {boolean}
   */
  hasOwnProperty: function hasOwnProperty(key) {
    return _constants.OBJECT_PROTOTYPE.hasOwnProperty.call(this, key);
  },


  /**
   * does this include the value passed
   *
   * @param {*} value
   * @returns {boolean}
   */
  includes: function includes(value) {
    var keys = this.keys();
    var cleanValue = (0, _crio.getCleanValue)(value);

    var index = -1,
        currentKey = void 0;

    while (++index < this.length) {
      currentKey = keys[index];

      if ((0, _isEqual2.default)((0, _crio.getCleanValue)(this[currentKey]), cleanValue)) {
        return true;
      }
    }

    return false;
  },


  /**
   * is the Crio a CrioArray
   *
   * @returns {boolean}
   */
  isArray: function isArray() {
    return (0, _is.isCrioArray)(this);
  },


  /**
   * is the Crio a CrioObject
   *
   * @returns {boolean}
   */
  isObject: function isObject() {
    return (0, _is.isCrioObject)(this);
  },


  /**
   * shallowly merge the objects passed with this
   *
   * @param {Array<Object>} objects
   * @returns {CrioObject}
   */
  merge: function merge() {
    for (var _len2 = arguments.length, objects = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
      objects[_key2] = arguments[_key2];
    }

    return mergeCrios.apply(undefined, [this].concat(objects));
  },


  /**
   * shallowly merge the objects passed with the deeply-nested location determined by keys
   *
   * @param {Array<string|number>} keys
   * @param {Array<Object>} objects
   * @returns {CrioObject}
   */
  mergeIn: function mergeIn(keys) {
    for (var _len3 = arguments.length, objects = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
      objects[_key3 - 1] = arguments[_key3];
    }

    (0, _crio.throwTypeErrorIfKeysInvalid)(keys);

    var length = keys.length;

    if (length === 1) {
      var _key4 = keys[0];

      if ((0, _is.isCrio)(this[_key4])) {
        return this.set(_key4, mergeCrios.apply(undefined, [this[_key4]].concat(objects)));
      }

      var object = objects[0];
      var restOfObjects = objects.slice(1);


      return this.set(_key4, mergeCrios.apply(undefined, [object].concat(_toConsumableArray(restOfObjects))));
    }

    if (!length) {
      return this;
    }

    var _keys2 = _toArray(keys);

    var key = _keys2[0];

    var restOfKeys = _keys2.slice(1);

    var plainObject = (0, _crio.getPlainObject)(this, false),
        isKeySet = false,
        isTargetKey = false;

    this.forEach(function (currentValue, currentKey) {
      isTargetKey = currentKey === key;

      if (isTargetKey) {
        isKeySet = true;

        plainObject[currentKey] = (0, _is.isCrio)(currentValue) ? currentValue.mergeIn.apply(currentValue, [restOfKeys].concat(objects)) : _loops.createDeeplyNestedObject.apply(undefined, [restOfKeys].concat(objects));
      } else {
        plainObject[currentKey] = currentValue;
      }
    });

    if (!isKeySet) {
      plainObject[key] = mergeCrios.apply(undefined, objects);
    }

    return (0, _crio.getSameCrioIfUnchanged)(this, plainObject);
  },


  /**
   * execute a function with the mutated value of this and return the re-crioed version
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {*}
   */
  mutate: function mutate(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var result = fn.call(thisArg, this.thaw(), this);
    var crioedValue = getCrioedValue(result);

    if ((0, _is.isCrio)(crioedValue)) {
      return this.equals(crioedValue) ? this : crioedValue;
    }

    return crioedValue;
  },


  /**
   * return new CrioArray of values in collection for the property method
   *
   * @param {string} key
   * @returns {CrioArray}
   */
  pluck: function pluck(key) {
    var array = [];

    this.forEach(function (value) {
      if (value.hasOwnProperty(key)) {
        array.push(value[key]);
      }
    });

    return new CrioArray(array);
  },


  /**
   * pluck the deeply-nested value based on keys
   *
   * @param {Array<number|string>} keys
   * @returns {Array<*>}
   */
  pluckIn: function pluckIn(keys) {
    (0, _crio.throwTypeErrorIfKeysInvalid)(keys);

    var length = keys.length;

    if (length === 1) {
      return this.pluck(keys[0]);
    }

    if (!length) {
      return this;
    }

    var _keys3 = _toArray(keys);

    var key = _keys3[0];

    var restOfKeys = _keys3.slice(1);

    var array = [];

    this.forEach(function (value) {
      if (value.hasOwnProperty(key) && (0, _is.isCrio)(value[key])) {
        var deepValue = value[key].getIn(restOfKeys);

        if (!(0, _isUndefined2.default)(deepValue)) {
          array.push(deepValue);
        }
      }
    });

    return new CrioArray(array);
  },


  /**
   * set key in this to be value
   *
   * @param {string|number} key
   * @param {*} value
   * @returns {CrioArray|CrioObject}
   */
  set: function set(key, value) {
    if (this[key] === value) {
      return this;
    }

    var shallowClone = (0, _is.isCrioArray)(this) ? (0, _loops.shallowCloneArray)(this) : (0, _loops.shallowCloneObject)(this);

    shallowClone[key] = value;

    return (0, _crio.getSameCrioIfUnchanged)(this, shallowClone);
  },


  /**
   * set deeply-nested value in this based on keys
   *
   * @param {Array<string|number>} keys
   * @param {number} keys.length
   * @param {*} value
   * @returns {CrioArray|CrioObject}
   */
  setIn: function setIn(keys, value) {
    (0, _crio.throwTypeErrorIfKeysInvalid)(keys);

    var length = keys.length;

    if (length === 1) {
      return this.set(keys[0], value);
    }

    if (!length) {
      return this;
    }

    var _keys4 = _toArray(keys);

    var key = _keys4[0];

    var restOfKeys = _keys4.slice(1);

    if (!this[key]) {
      return this.set(key, (0, _loops.createDeeplyNestedObject)(restOfKeys, value));
    }

    var plainObject = (0, _crio.getPlainObject)(this, false);

    this.forEach(function (currentValue, currentKey) {
      plainObject[currentKey] = (0, _crio.getDeeplyNestedValue)(currentValue, value, currentKey === key, restOfKeys);
    });

    return (0, _crio.getSameCrioIfUnchanged)(this, plainObject);
  },


  /**
   * return the non-crio version of the object
   *
   * @returns {Array<*>|Object}
   */
  thaw: function thaw() {
    var plainObject = (0, _crio.getPlainObject)(this, false);

    this.forEach(function (value, key) {
      plainObject[key] = (0, _is.isCrio)(value) ? value.thaw() : value;
    });

    return plainObject;
  },


  /**
   * convert this to a CrioArray
   *
   * @returns {CrioArray}
   */
  toArray: function toArray() {
    if ((0, _is.isCrioArray)(this)) {
      return this;
    }

    return this.reduce(function (array, value) {
      return array.concat([value]);
    }, []);
  },


  /**
   * get the stringified version of this
   *
   * @returns {string}
   */
  toLocaleString: function toLocaleString() {
    return (0, _stringify2.default)(this);
  },


  /**
   * convert this to a CrioObject
   *
   * @returns {CrioObject}
   */
  toObject: function toObject() {
    if ((0, _is.isCrioObject)(this)) {
      return this;
    }

    return this.reduce(function (object, value, index) {
      object[index] = value;

      return object;
    }, {});
  },


  /**
   * get the stringified version of this
   *
   * @returns {string}
   */
  toString: function toString() {
    return (0, _stringify2.default)(this);
  },


  /**
   * get the valueOf for this
   *
   * @returns {CrioArray|CrioObject}
   */
  valueOf: function valueOf() {
    return this;
  },


  /**
   * get the values for this
   *
   * @returns {Array<*>}
   */
  values: function values() {
    return _constants.OBJECT.values(this);
  }
};

Crio.prototype = (0, _constants.OBJECT_CREATE)(null, (0, _crio.createPrototypeObject)(CRIO_PROTOTYPE));

var CrioArray = function (_Crio) {
  _inherits(CrioArray, _Crio);

  /**
   * create CrioArray class extending Crio with built prototype
   *
   * @param {Array<*>} array
   * @param {string} hashCode
   */
  function CrioArray(array, hashCode) {
    _classCallCheck(this, CrioArray);

    return _possibleConstructorReturn(this, (CrioArray.__proto__ || Object.getPrototypeOf(CrioArray)).call(this, array, hashCode));
  }

  return CrioArray;
}(Crio);

var CRIO_ARRAY_PROTOTYPE = (_CRIO_ARRAY_PROTOTYPE = {
  /**
   * concatenate the arguments passed with the current array
   *
   * @param {Array<*> } args
   * @returns {CrioArray}
   */
  concat: function concat() {
    for (var _len4 = arguments.length, args = Array(_len4), _key5 = 0; _key5 < _len4; _key5++) {
      args[_key5] = arguments[_key5];
    }

    if (!args.length) {
      return this;
    }

    var shallowClone = (0, _loops.shallowCloneArray)(this);

    return (0, _crio.getSameCrioIfUnchanged)(this, _constants.ARRAY_PROTOTYPE.concat.apply(shallowClone, args));
  },


  constructor: CrioArray,

  /**
   * return a new array with the appropriate arguments for copyWithin applied
   *
   * @param {Array<*>} args
   * @returns {CrioArray}
   */
  copyWithin: function copyWithin() {
    for (var _len5 = arguments.length, args = Array(_len5), _key6 = 0; _key6 < _len5; _key6++) {
      args[_key6] = arguments[_key6];
    }

    if (!args.length) {
      return this;
    }

    var shallowClone = (0, _loops.shallowCloneArray)(this);
    var copiedArray = _constants.ARRAY_PROTOTYPE.copyWithin.apply(shallowClone, args);

    return (0, _crio.getSameCrioIfUnchanged)(this, copiedArray);
  },


  /**
   * find the values in this that do not exist in any of the arrays passed
   *
   * @param {Array<Array>} arrays
   * @returns {CrioArray}
   */
  difference: function difference() {
    for (var _len6 = arguments.length, arrays = Array(_len6), _key7 = 0; _key7 < _len6; _key7++) {
      arrays[_key7] = arguments[_key7];
    }

    if (!arrays.length) {
      return this;
    }

    var indexOfValue = void 0;

    var shallowClone = arrays.reduce(function (differenceArray, array) {
      if ((0, _isArray2.default)(array) || (0, _is.isCrioArray)(array)) {
        (0, _forEach2.default)(array, function (value) {
          indexOfValue = differenceArray.indexOf(value);

          if (!!~indexOfValue) {
            differenceArray.splice(indexOfValue, 1);
          }
        });
      }

      return differenceArray;
    }, (0, _loops.shallowCloneArray)(this));

    return (0, _crio.getSameCrioIfUnchanged)(this, shallowClone);
  },


  /**
   * does every item in this match the result of fn
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {boolean}
   */
  every: function every(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    return _constants.ARRAY_PROTOTYPE.every.call(this, fn, thisArg);
  },


  /**
   * return a new array with the appropriate arguments for fill applied
   *
   * @param {Array<*>} args
   * @returns {CrioArray}
   */
  fill: function fill() {
    for (var _len7 = arguments.length, args = Array(_len7), _key8 = 0; _key8 < _len7; _key8++) {
      args[_key8] = arguments[_key8];
    }

    if (!args.length) {
      return this;
    }

    var shallowClone = (0, _loops.shallowCloneArray)(this);
    var filledArray = _constants.ARRAY_PROTOTYPE.fill.apply(shallowClone, args);

    return (0, _crio.getSameCrioIfUnchanged)(this, filledArray);
  },


  /**
   * filter this based on truthy results from fn
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {CrioArray}
   */
  filter: function filter(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var filteredArray = _constants.ARRAY_PROTOTYPE.filter.call(this, fn, thisArg);

    return (0, _crio.getSameCrioIfUnchanged)(this, filteredArray);
  },


  /**
   * find the first item that returns truthy for fn
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {*}
   */
  find: function find(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    return _constants.ARRAY_PROTOTYPE.find.call(this, fn, thisArg);
  },


  /**
   * find the index of the first item that returns truthy for fn
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {number}
   */
  findIndex: function findIndex(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    return _constants.ARRAY_PROTOTYPE.findIndex.call(this, fn, thisArg);
  },


  /**
   * return the first X number of items, based on number
   *
   * @param {number} number
   * @returns {CrioArray}
   */
  first: function first() {
    var number = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 1;

    if (number >= this.length) {
      return this;
    }

    return this.slice(0, number);
  },


  /**
   * loop over this, executing fn
   *
   * @param {function} fn
   * @param {*} thisArg
   */
  forEach: function forEach(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    _constants.ARRAY_PROTOTYPE.forEach.call(this, fn, thisArg);
  },


  /**
   * if the index of the value passed exists, return the
   * first instance of it, else return -1
   *
   * @param {*} value
   * @returns {number}
   */
  indexOf: function indexOf(value) {
    return _constants.ARRAY_PROTOTYPE.indexOf.call(this, value);
  },


  /**
   * find the values in that exist in this and each of the arrays passed
   *
   * @param {Array<Array>} arrays
   * @returns {CrioArray}
   */
  intersection: function intersection() {
    for (var _len8 = arguments.length, arrays = Array(_len8), _key9 = 0; _key9 < _len8; _key9++) {
      arrays[_key9] = arguments[_key9];
    }

    if (!arrays.length) {
      return this;
    }

    var allArrays = [this].concat(arrays);
    var allArraysLength = allArrays.length;

    var indices = [],
        indexOfValue = void 0;

    var intersectingValues = allArrays.reduce(function (values, array) {
      if ((0, _isArray2.default)(array) || (0, _is.isCrioArray)(array)) {
        (0, _forEach2.default)(array, function (value) {
          indexOfValue = values.indexOf(value);

          if (!!~indexOfValue) {
            indices[indexOfValue]++;
          } else {
            indices[values.length] = 1;
            values.push(value);
          }
        });
      }

      return values;
    }, []).filter(function (value, index) {
      return indices[index] === allArraysLength;
    });

    return (0, _crio.getSameCrioIfUnchanged)(this, intersectingValues);
  },


  /**
   * combine the values in this, with separator as the separator
   *
   * @param {string} separator
   * @returns {string}
   */
  join: function join() {
    var separator = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : ',';

    return _constants.ARRAY_PROTOTYPE.join.call(this, separator);
  },


  /**
   * return the keys of this
   *
   * @returns {Array<string>}
   */
  keys: function keys() {
    return (0, _constants.OBJECT_KEYS)(this).map(_loops.convertToNumber);
  },


  /**
   * return the last X number of items, based on number
   *
   * @param {number} number
   * @returns {CrioArray}
   */
  last: function last() {
    var number = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 1;

    if (number >= this.length) {
      return this;
    }

    return this.slice(this.length - number, this.length);
  },


  /**
   * if the index of the value passed exists, return the
   * last instance of it, else return -1
   *
   * @param {*} value
   * @returns {number}
   */
  lastIndexOf: function lastIndexOf(value) {
    return _constants.ARRAY_PROTOTYPE.lastIndexOf.call(this, value);
  },


  /**
   * return the values mapped by fn as a new CrioArray
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {CrioArray}
   */
  map: function map(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var mappedArray = _constants.ARRAY_PROTOTYPE.map.call(this, fn, thisArg);

    return (0, _crio.getSameCrioIfUnchanged)(this, mappedArray);
  },


  /**
   * return a new CrioArray with the last item removed
   *
   * @returns {CrioArray}
   */
  pop: function pop() {
    return this.slice(0, this.length - 1);
  },


  /**
   * add items to the current CrioArray
   *
   * @param {Array<*>} items
   * @returns {CrioArray}
   */
  push: function push() {
    for (var _len9 = arguments.length, items = Array(_len9), _key10 = 0; _key10 < _len9; _key10++) {
      items[_key10] = arguments[_key10];
    }

    if (!items.length) {
      return this;
    }

    return this.concat(items);
  },


  /**
   * reduce the values in the array based on starting with defaultValue
   *
   * @param {function} fn
   * @param {*} defaultValue
   * @param {*} thisArg
   * @returns {*}
   */
  reduce: function reduce(fn, defaultValue) {
    var thisArg = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : this;

    var reducedValue = _constants.ARRAY_PROTOTYPE.reduce.call(this, fn, defaultValue, thisArg);
    var crioedValue = getCrioedValue(reducedValue);

    return (0, _is.isCrio)(crioedValue) && this.equals(crioedValue) ? this : crioedValue;
  },


  /**
   * reduce the values in the array based on starting with defaultValue,
   * but starting from the end and working to the beginning
   *
   * @param {function} fn
   * @param {*} defaultValue
   * @param {*} thisArg
   * @returns {*}
   */
  reduceRight: function reduceRight(fn, defaultValue) {
    var thisArg = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : this;

    var reducedValue = _constants.ARRAY_PROTOTYPE.reduceRight.call(this, fn, defaultValue, thisArg);
    var crioedValue = getCrioedValue(reducedValue);

    return (0, _is.isCrio)(crioedValue) && this.equals(crioedValue) ? this : crioedValue;
  },


  /**
   * reverse the order of the CrioArray
   *
   * @returns {CrioArray}
   */
  reverse: function reverse() {
    var newArray = [];

    (0, _forEachRight2.default)(this, function (value) {
      newArray.push(value);
    });

    return (0, _crio.getSameCrioIfUnchanged)(this, newArray);
  },


  /**
   * return the CrioArray with the first item removed
   *
   * @returns {CrioArray}
   */
  shift: function shift() {
    return this.slice(1);
  },


  /**
   * return the sliced version of the current CrioArray
   *
   * @param {Array<*>} args
   * @returns {CrioArray}
   */
  slice: function slice() {
    for (var _len10 = arguments.length, args = Array(_len10), _key11 = 0; _key11 < _len10; _key11++) {
      args[_key11] = arguments[_key11];
    }

    if (!args.length) {
      return this;
    }

    var slicedArray = _constants.ARRAY_PROTOTYPE.slice.apply(this, args);

    return (0, _crio.getSameCrioIfUnchanged)(this, slicedArray);
  },


  /**
   * does this return truthy for at least one of the returns of fn
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {boolean}
   */
  some: function some(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    return _constants.ARRAY_PROTOTYPE.some.call(this, fn, thisArg);
  },


  /**
   * return a sorted version of the current CrioArray
   *
   * @param {function} fn
   * @returns {CrioArray}
   */
  sort: function sort(fn) {
    var shallowClone = (0, _loops.shallowCloneArray)(this);

    _constants.ARRAY_PROTOTYPE.sort.call(shallowClone, fn);

    return (0, _crio.getSameCrioIfUnchanged)(this, shallowClone);
  },


  /**
   * return the spliced version of the current CrioArray
   *
   * @param {Array<*>} args
   * @returns {CrioArray}
   */
  splice: function splice() {
    for (var _len11 = arguments.length, args = Array(_len11), _key12 = 0; _key12 < _len11; _key12++) {
      args[_key12] = arguments[_key12];
    }

    if (!args.length) {
      return this;
    }

    var shallowClone = (0, _loops.shallowCloneArray)(this);

    _constants.ARRAY_PROTOTYPE.splice.apply(shallowClone, args);

    return (0, _crio.getSameCrioIfUnchanged)(this, shallowClone);
  },


  /**
   * return the current CrioArray with the duplicate values removed
   *
   * @returns {CrioArray}
   */
  unique: function unique() {
    var hashArray = [],
        newArray = [],
        hasHashCode = false,
        hashCode = void 0;

    var filteredCrioArray = this.filter(function (value) {
      hashCode = value[_constants.CRIO_HASH_CODE];
      hasHashCode = !(0, _isUndefined2.default)(hashCode);

      if (!newArray.includes(value) && (!hasHashCode || !hashArray.includes(hashCode))) {
        newArray.push(value);

        if (hasHashCode) {
          hashArray.push(hashCode);
        }

        return true;
      }

      return false;
    });

    return (0, _crio.getSameCrioIfUnchanged)(this, filteredCrioArray);
  },


  /**
   * add the args passed to the current CrioArray
   *
   * @param {Array<*>} args
   * @returns {CrioArray}
   */
  unshift: function unshift() {
    for (var _len12 = arguments.length, args = Array(_len12), _key13 = 0; _key13 < _len12; _key13++) {
      args[_key13] = arguments[_key13];
    }

    if (!args.length) {
      return this;
    }

    var unshiftedArray = _constants.ARRAY_PROTOTYPE.concat.apply(args, this);

    return (0, _crio.getSameCrioIfUnchanged)(this, unshiftedArray);
  },


  /**
   * find the values that are the symmetric difference of this and the arrays passed
   *
   * @param {Array<Array>} arrays
   * @returns {CrioArray}
   */
  xor: function xor() {
    for (var _len13 = arguments.length, arrays = Array(_len13), _key14 = 0; _key14 < _len13; _key14++) {
      arrays[_key14] = arguments[_key14];
    }

    if (!arrays.length) {
      return this;
    }

    var allArrays = [this].concat(arrays);

    var indicesToRemove = [],
        indexOfValue = void 0;

    var xorValues = allArrays.reduce(function (values, array) {
      if ((0, _isArray2.default)(array) || (0, _is.isCrioArray)(array)) {
        (0, _forEach2.default)(array, function (value) {
          indexOfValue = values.indexOf(value);

          if (!!~indexOfValue) {
            indicesToRemove.push(indexOfValue);
          } else {
            values.push(value);
          }
        });
      }

      return values;
    }, []).filter(function (value, index) {
      return !~indicesToRemove.indexOf(index);
    });

    return (0, _crio.getSameCrioIfUnchanged)(this, xorValues);
  }
}, _defineProperty(_CRIO_ARRAY_PROTOTYPE, _constants.CRIO_TYPE, _constants.CRIO_ARRAY), _defineProperty(_CRIO_ARRAY_PROTOTYPE, Symbol.iterator, _constants.ARRAY_PROTOTYPE[Symbol.iterator]), _CRIO_ARRAY_PROTOTYPE);

CrioArray.prototype = (0, _constants.OBJECT_CREATE)(Crio.prototype, (0, _crio.createPrototypeObject)(CRIO_ARRAY_PROTOTYPE));

var CrioObject = function (_Crio2) {
  _inherits(CrioObject, _Crio2);

  /**
   * create CrioObject class extending Crio with built prototype
   *
   * @param {Object} object
   * @param {string} hashCode
   */
  function CrioObject(object, hashCode) {
    _classCallCheck(this, CrioObject);

    return _possibleConstructorReturn(this, (CrioObject.__proto__ || Object.getPrototypeOf(CrioObject)).call(this, object, hashCode));
  }

  return CrioObject;
}(Crio);

var CRIO_OBJECT_PROTOTYPE = (_CRIO_OBJECT_PROTOTYP = {
  constructor: CrioObject,

  /**
   * loop over the object and if all fn calls returns true, then return true
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {boolean}
   */
  every: function every(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var keys = this.keys();

    var index = -1,
        key = void 0;

    while (++index < this.length) {
      key = keys[index];

      if (!fn.call(thisArg, this[key], key, this)) {
        return false;
      }
    }

    return true;
  },


  /**
   * filter the current CrioArray by the truthy return of fn
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {CrioObject}
   */
  filter: function filter(fn) {
    var _this6 = this;

    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var value = void 0;

    var newObject = this.keys().reduce(function (object, key) {
      value = _this6[key];

      if (fn.call(thisArg, value, key, _this6)) {
        object[key] = value;
      }

      return object;
    }, {});

    return (0, _crio.getSameCrioIfUnchanged)(this, newObject);
  },


  /**
   * find the value in this that yields a truthy return from fn
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {*}
   */
  find: function find(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var keys = this.keys();

    var index = -1,
        key = void 0;

    while (++index < this.length) {
      key = keys[index];

      if (fn.call(thisArg, this[key], key, this)) {
        return this[key];
      }
    }

    return undefined;
  },


  /**
   * find the key in this that yields a truthy return from fn
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {*}
   */
  findKey: function findKey(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var keys = this.keys();

    var index = -1,
        key = void 0;

    while (++index < this.length) {
      key = keys[index];

      if (fn.call(thisArg, this[key], key, this)) {
        return key;
      }
    }

    return undefined;
  },


  /**
   * loop over this, executing fn
   *
   * @param {function} fn
   * @param {*} thisArg
   */
  forEach: function forEach(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    (0, _loops.forEachObject)(this, fn, thisArg);
  },


  /**
   * is this the prototype of the object passed
   *
   * @param {*} object
   * @returns {boolean}
   */
  isPrototypeOf: function isPrototypeOf(object) {
    return _constants.OBJECT_PROTOTYPE.isPrototypeOf.call(this, object);
  },


  /**
   * get the keys of this
   *
   * @returns {Array<string>}
   */
  keys: function keys() {
    return (0, _constants.OBJECT_KEYS)(this);
  },


  /**
   * return the new object based on the mapped values of this
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {CrioObject}
   */
  map: function map(fn) {
    var _this7 = this;

    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var value = void 0,
        result = void 0;

    var newObject = this.keys().reduce(function (object, key) {
      value = _this7[key];
      result = fn.call(thisArg, value, key, _this7);

      object[key] = getCrioedValue(result);

      return object;
    }, {});

    return (0, _crio.getSameCrioIfUnchanged)(this, newObject);
  },


  /**
   * is the property passed enumerable
   *
   * @param {string} property
   * @returns {boolean}
   */
  propertyIsEnumerable: function propertyIsEnumerable(property) {
    return _constants.OBJECT_PROTOTYPE.propertyIsEnumerable.call(this, property);
  },


  /**
   * reduce the values in the object based on starting with defaultValue
   *
   * @param {function} fn
   * @param {*} defaultValue
   * @param {*} thisArg
   * @returns {*}
   */
  reduce: function reduce(fn, defaultValue) {
    var _this8 = this;

    var thisArg = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : this;

    var reducedValue = this.keys().reduce(function (accumulation, key) {
      return fn.call(thisArg, accumulation, _this8[key], key, _this8);
    }, defaultValue);
    var crioedValue = getCrioedValue(reducedValue);

    return (0, _is.isCrio)(crioedValue) && this.equals(crioedValue) ? this : crioedValue;
  },


  /**
   * reduce the values in the array based on starting with defaultValue,
   * but starting from the end and working to the beginning
   *
   * @param {function} fn
   * @param {*} defaultValue
   * @param {*} thisArg
   * @returns {*}
   */
  reduceRight: function reduceRight(fn, defaultValue) {
    var _this9 = this;

    var thisArg = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : this;

    var reducedValue = this.keys().reduce(function (accumulation, key) {
      return fn.call(thisArg, accumulation, _this9[key], key, _this9);
    }, defaultValue);
    var crioedValue = getCrioedValue(reducedValue);

    return (0, _is.isCrio)(crioedValue) && this.equals(crioedValue) ? this : crioedValue;
  },


  /**
   * loop over the object and if the fn passed returns true, then return true
   *
   * @param {function} fn
   * @param {*} thisArg
   * @returns {boolean}
   */
  some: function some(fn) {
    var thisArg = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this;

    var keys = this.keys();

    var index = -1,
        key = void 0;

    while (++index < this.length) {
      key = keys[index];

      if (fn.call(thisArg, this[key], key, this)) {
        return true;
      }
    }

    return false;
  }
}, _defineProperty(_CRIO_OBJECT_PROTOTYP, _constants.CRIO_TYPE, _constants.CRIO_OBJECT), _defineProperty(_CRIO_OBJECT_PROTOTYP, Symbol.iterator, function () {
  var _this10 = this;

  var index = 0,
      key = void 0,
      value = void 0;

  return {
    next: function next() {
      value = _this10[key];

      if (index < _this10.length) {
        index++;

        return {
          done: false,
          value: value
        };
      } else {
        return {
          done: true
        };
      }
    }
  };
}), _CRIO_OBJECT_PROTOTYP);

CrioObject.prototype = (0, _constants.OBJECT_CREATE)(Crio.prototype, (0, _crio.createPrototypeObject)(CRIO_OBJECT_PROTOTYPE));

exports.getCrioedValue = getCrioedValue;
exports.mergeArrays = mergeArrays;
exports.mergeCrios = mergeCrios;
exports.CrioArray = CrioArray;
exports.CrioObject = CrioObject;