'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

require('core-js/modules/es6.array.copy-within');

require('core-js/modules/es6.array.every');

require('core-js/modules/es6.array.fill');

require('core-js/modules/es6.array.find');

require('core-js/modules/es6.array.find-index');

require('core-js/modules/es6.array.iterator');

require('core-js/modules/es6.array.some');

require('core-js/modules/es7.array.includes');

require('core-js/modules/es7.object.enumerable-entries');

require('core-js/modules/es7.object.enumerable-keys');

require('core-js/modules/es7.object.enumerable-values');

var _isArray = require('lodash/isArray');

var _isArray2 = _interopRequireDefault(_isArray);

var _isPlainObject = require('lodash/isPlainObject');

var _isPlainObject2 = _interopRequireDefault(_isPlainObject);

var _classes = require('./classes');

var _is = require('./utils/is');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// external dependencies


// ESNext
var throwTypeError = function throwTypeError(type) {
  throw new TypeError('Must pass ' + type + ' to crio.' + type + '.');
};

/**
 * generate a new CrioArray or CrioObject
 *
 * @param {*} object
 * @returns {CrioArray|CrioObject|Array<*>|Object|*}
 */
// ES2015
var createCrio = function createCrio() {
  var object = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  return (0, _classes.getCrioedValue)(object);
};

/**
 * create a new CrioArray
 *
 * @param {Array<*>} array
 * @returns {CrioArray}
 */
createCrio.array = function () {
  var array = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];

  if (!(0, _isArray2.default)(array)) {
    throwTypeError('array');
  }

  return new _classes.CrioArray(array);
};

createCrio.isArray = _is.isCrioArray;
createCrio.isCrio = _is.isCrio;
createCrio.isObject = _is.isCrioObject;

/**
 * create a new CrioObject
 *
 * @param {Object} object
 * @returns {CrioObject}
 */
createCrio.object = function () {
  var object = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  if (!(0, _isPlainObject2.default)(object)) {
    throwTypeError('object');
  }

  return new _classes.CrioObject(object);
};

exports.default = createCrio;
module.exports = exports['default'];