'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.isReactElement = exports.isCrioObject = exports.isCrioArray = exports.isCrio = undefined;

var _constants = require('./constants');

/**
 * determine if object passed is a Crio object
 *
 * @param {*} object
 * @returns {boolean}
 */
var isCrio = function isCrio(object) {
  return !!(object && object[_constants.CRIO_TYPE]);
};

/**
 * determine if object passed is a CrioArray
 *
 * @param {*} object
 * @returns {boolean}
 */
// constants
var isCrioArray = function isCrioArray(object) {
  return isCrio(object) && object[_constants.CRIO_TYPE] === _constants.CRIO_ARRAY;
};

/**
 * determine if object passed is a CrioObject
 *
 * @param {*} object
 * @returns {boolean}
 */
var isCrioObject = function isCrioObject(object) {
  return isCrio(object) && object[_constants.CRIO_TYPE] === _constants.CRIO_OBJECT;
};

/**
 * determine if object is a React element
 *
 * @param {any} object
 * @param {string|Symbol} object.$$typeof
 * @return {boolean}
 */
var isReactElement = function isReactElement(object) {
  return !!object && object.$$typeof === _constants.REACT_ELEMENT_TYPE;
};

exports.isCrio = isCrio;
exports.isCrioArray = isCrioArray;
exports.isCrioObject = isCrioObject;
exports.isReactElement = isReactElement;