'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.shallowCloneObject = exports.shallowCloneArray = exports.mergeObjects = exports.forEachObject = exports.createDeeplyNestedObject = exports.convertToNumber = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _forEach = require('lodash/forEach');

var _forEach2 = _interopRequireDefault(_forEach);

var _isNumber = require('lodash/isNumber');

var _isNumber2 = _interopRequireDefault(_isNumber);

var _isPlainObject = require('lodash/isPlainObject');

var _isPlainObject2 = _interopRequireDefault(_isPlainObject);

var _constants = require('./constants');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _toArray(arr) { return Array.isArray(arr) ? arr : Array.from(arr); } // external dependencies


/**
 * convert the value passed into its numeric form
 *
 * @param {*} value
 * @returns {number}
 */
var convertToNumber = function convertToNumber(value) {
  return +value;
};

/**
 * forEach loop specific to objects
 *
 * @param {CrioObject} crio
 * @param {Function} fn
 * @param {*} thisArg
 */
var forEachObject = function forEachObject(crio, fn, thisArg) {
  (0, _forEach2.default)(crio.keys(), function (key) {
    fn.call(thisArg, crio[key], key, crio);
  });
};

/**
 * create a deeply-nested new object with value at last key location
 *
 * @param {Array<string|number>} keys
 * @param {*} value
 * @returns {Array<*>|object}
 */
var createDeeplyNestedObject = function createDeeplyNestedObject(keys, value) {
  var _keys = _toArray(keys);

  var key = _keys[0];

  var restOfKeys = _keys.slice(1);

  var valueToSave = restOfKeys.length ? createDeeplyNestedObject(restOfKeys, value) : value;

  return (0, _isNumber2.default)(key) ? [valueToSave] : _defineProperty({}, key, valueToSave);
};

/**
 * shallowly merge source objects into target object
 *
 * @param {Object} target
 * @param {Array<Object>} sources
 * @returns {Array<*>|Object}
 */
var mergeObjects = function mergeObjects(target, sources) {
  return sources.reduce(function (plainObject, source) {
    if ((0, _isPlainObject2.default)(source)) {
      plainObject = _extends({}, plainObject, source);
    }

    return plainObject;
  }, _extends({}, target));
};

/**
 * shallowly clone an array
 *
 * @param {Array<*>} array
 * @returns {Array<T>}
 */
var shallowCloneArray = function shallowCloneArray(array) {
  return _constants.ARRAY_PROTOTYPE.slice.call(array, 0, array.length);
};

/**
 * shallowly clone an object
 *
 * @param {Object} object
 * @returns {Object}
 */
var shallowCloneObject = function shallowCloneObject(object) {
  return _extends({}, object);
};

exports.convertToNumber = convertToNumber;
exports.createDeeplyNestedObject = createDeeplyNestedObject;
exports.forEachObject = forEachObject;
exports.mergeObjects = mergeObjects;
exports.shallowCloneArray = shallowCloneArray;
exports.shallowCloneObject = shallowCloneObject;